"""
Export a single recipe to Paprika App

file format:
    1. export recipe to json
    2. compress file as gz: naming convention: a_recipe_name.paprikarecipe (Singular)
    3. zip this file as some_recipes.paprikarecipes (Plural!)
"""

import base64
import glob
import gzip
import json
import logging
import os
import re
import secrets
import tempfile
from datetime import datetime
from pathlib import Path
from typing import Any

import httpx
from jinja2 import Template

from kptncook.config import settings
from kptncook.exporter_utils import (
    asciify_string,
    get_cover,
    move_to_target_dir,
    write_zip,
)
from kptncook.ingredient_groups import iter_ingredient_groups
from kptncook.models import Image, Ingredient, Recipe, localized_fallback

PAPRIKA_RECIPE_TEMPLATE = """{
   "uid":"{{recipe.id.oid}}",
   "name":"{{localized_fallback(recipe.localized_title)|default('',true)}}",
   "directions": "{% for step in recipe.steps %}{{localized_fallback(step.title)|default('',true)}}\\n{% endfor %}",
   "servings":"2",
   "rating":0,
   "difficulty":"",
   "ingredients":"{{ingredients_text}}",
   "notes":"",
   "created":"{{dtnow}}",
   "image_url":null,
   "cook_time":"{{recipe.cooking_time|default('',true)}}",
   "prep_time":"{{recipe.preparation_time|default('',true)}}",
   "source":"Kptncook",
   "source_url":"",
   "hash" : "{{hash}}",
   "photo_hash":null,
   "photos":[],
   "photo": "{{cover_filename}}",
   "nutritional_info":"{% for nutrient, amount in recipe.recipe_nutrition %}{{nutrient}}: {{amount}}\\n{% endfor %}",
   "photo_data":"{{cover_img}}",
   "photo_large":null,
   "categories":["Kptncook"]
}
"""  # noqa: E501

logger = logging.getLogger(__name__)


class GeneratedData:
    def __init__(
        self, cover_filename: str | None, cover_img: str | None, dtnow: str, hash_: str
    ):
        self.cover_filename = cover_filename
        self.cover_img = cover_img
        self.dtnow = dtnow
        self.hash = hash_


class PaprikaExporter:
    invalid_control_chars = re.compile(r"[\x00-\x08\x0B-\x0C\x0E-\x1F\x7F-\x9F]")
    template = Template(PAPRIKA_RECIPE_TEMPLATE, trim_blocks=True)
    unescaped_newline = re.compile(r"(?<!\\)\n")

    def export(self, recipes: list[Recipe]) -> str:
        export_data = self.get_export_data(recipes=recipes)
        filename = self.get_export_filename(export_data=export_data, recipes=recipes)
        tmp_dir = tempfile.mkdtemp()
        filename_full_path = self.save_recipes(
            export_data=export_data, directory=tmp_dir, filename=filename
        )
        move_to_target_dir(
            source=filename_full_path, target=os.path.join(str(Path.cwd()), filename)
        )
        return filename

    def get_export_filename(
        self, export_data: dict[str, str], recipes: list[Recipe]
    ) -> str:
        if len(export_data) == 1:
            return (
                asciify_string(
                    s=localized_fallback(recipes[0].localized_title) or "recipe"
                )
                + ".paprikarecipes"
            )
        else:
            return "allrecipes.paprikarecipes"

    def get_generated_data(self, recipe: Recipe) -> GeneratedData:
        """Just to make testing easier and only have one method to mock in tests."""
        cover_filename, cover_img = self.get_cover_img_as_base64_string(recipe=recipe)
        dtnow = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        hash = secrets.token_hex(32)
        return GeneratedData(cover_filename, cover_img, dtnow, hash)

    def get_recipe_as_json_string(self, recipe: Recipe) -> str:
        generated = self.get_generated_data(recipe=recipe)
        ingredients_text = self.get_ingredients_text(recipe.ingredients)
        recipe_as_json = self.template.render(
            recipe=recipe,
            localized_fallback=localized_fallback,
            dtnow=generated.dtnow,
            cover_filename=generated.cover_filename,
            hash=generated.hash,
            cover_img=generated.cover_img,
            ingredients_text=ingredients_text,
        )
        recipe_as_json = self.invalid_control_chars.sub("", recipe_as_json)
        recipe_as_json = self.unescaped_newline.sub(" ", recipe_as_json)
        json.loads(recipe_as_json)  # check if valid json
        return recipe_as_json

    def get_ingredients_text(self, ingredients: list[Ingredient]) -> str:
        lines = []
        for group_label, group_ingredients in iter_ingredient_groups(ingredients):
            if group_label:
                lines.append(f"{group_label}:")
            for ingredient in group_ingredients:
                line = self.format_ingredient_line(ingredient)
                if line:
                    lines.append(line)
        if not lines:
            return ""
        return "\\n".join(lines) + "\\n"

    def format_ingredient_line(self, ingredient: Ingredient) -> str:
        parts: list[str] = []
        if ingredient.quantity:
            parts.append("{0:g}".format(ingredient.quantity))
        if ingredient.measure:
            parts.append(ingredient.measure)
        ingredient_name = (
            localized_fallback(ingredient.ingredient.uncountable_title) or ""
        )
        if ingredient_name:
            parts.append(ingredient_name)
        return " ".join(part for part in parts if part).strip()

    def get_export_data(self, recipes: list[Recipe]) -> dict[str, str]:
        export_data = dict()
        for recipe in recipes:
            try:
                recipe_as_json = self.get_recipe_as_json_string(recipe=recipe)
                export_data[str(recipe.id.oid)] = recipe_as_json
            except json.JSONDecodeError as e:
                logger.warning("Could not parse recipe %s: %s", recipe.id.oid, e)
        return export_data

    def save_recipes(
        self, export_data: dict[str, Any], filename: str, directory: str
    ) -> str:
        for id, recipe_as_json in export_data.items():
            recipe_as_gz = os.path.join(directory, "recipe_" + id + ".paprikarecipe")
            with gzip.open(recipe_as_gz, "wb") as f:
                f.write(recipe_as_json.encode("utf-8"))
        filename_full_path = os.path.join(directory, filename)
        gz_files = glob.glob(os.path.join(directory, "*.paprikarecipe"))
        logger.debug("Paprika export files: %s", gz_files)
        entries = [(Path(gz_file).name, Path(gz_file)) for gz_file in gz_files]
        write_zip(Path(filename_full_path), entries)
        return filename_full_path

    def get_cover_img_as_base64_string(
        self, recipe: Recipe
    ) -> tuple[str | None, str | None]:
        cover = get_cover(image_list=recipe.image_list)
        if cover is None:
            raise ValueError("No cover image found")
        cover_url = recipe.get_image_url(api_key=settings.kptncook_api_key)
        if not isinstance(cover_url, str):
            raise ValueError("Cover URL must be a string")
        try:
            response = httpx.get(cover_url)
            response.raise_for_status()
        except httpx.HTTPStatusError as exc:
            if exc.response.status_code == 404:
                title = localized_fallback(recipe.localized_title) or "kptncook-recipe"
                logger.warning('Cover image for "%s" not found online any more.', title)
            else:
                logger.error(
                    "HTTP error while fetching cover image (%s): %s",
                    exc.response.status_code,
                    exc,
                )
            return None, None
        return cover.name, base64.b64encode(response.content).decode("utf-8")

    def asciify_string(self, s: str) -> str:
        return asciify_string(s)

    def get_cover(self, image_list: list[Image] | None) -> Image | None:
        if not isinstance(image_list, list):
            raise ValueError("Parameter image_list must be a list")
        return get_cover(image_list)



"""
Password manager integration for retrieving credentials.
"""

import subprocess

from rich import print as rprint


def get_credential_from_command(command: str) -> str | None:
    """
    Execute a shell command to retrieve a credential from a password manager.

    Args:
        command: Shell command to execute (e.g., "op read op://vault/item/field")

    Returns:
        The credential as a string, or None if the command fails
    """
    try:
        result = subprocess.run(
            command, shell=True, capture_output=True, text=True, check=True
        )
        return result.stdout.strip()
    except subprocess.CalledProcessError as e:
        rprint(f"[red]Error executing password manager command: {e}[/red]")
        rprint(f"[red]Command output: {e.stderr}[/red]")
        return None
    except Exception as e:
        rprint(f"[red]Unexpected error: {e}[/red]")
        return None


def get_credentials(
    username_command: str | None = None,
    password_command: str | None = None,
    interactive_fallback: bool = True,
) -> tuple[str | None, str | None]:
    """
    Get username and password from password manager or interactive prompt.

    Args:
        username_command: Command to retrieve username
        password_command: Command to retrieve password
        interactive_fallback: Whether to fall back to interactive prompt

    Returns:
        Tuple of (username, password) or (None, None) if retrieval fails
    """
    username = None
    password = None

    # Try to get credentials from password manager
    if username_command:
        username = get_credential_from_command(username_command)
        if username:
            rprint("[green]✓ Username retrieved from password manager[/green]")

    if password_command:
        password = get_credential_from_command(password_command)
        if password:
            rprint("[green]✓ Password retrieved from password manager[/green]")

    # Fall back to interactive prompt if needed
    if interactive_fallback:
        if not username:
            from rich.prompt import Prompt

            username = Prompt.ask("Enter your kptncook email address")

        if not password:
            from rich.prompt import Prompt

            password = Prompt.ask("Enter your kptncook password", password=True)

    return username, password



import re
import shutil
import zipfile
from pathlib import Path
from typing import Iterable

from unidecode import unidecode

from kptncook.models import Image

ZipContent = bytes | str | Path


def asciify_string(s: str) -> str:
    s = unidecode(s)
    s = re.sub(r"[^\w\s]", "_", s)
    s = re.sub(r"\s+", "_", s)
    return s


def get_cover(image_list: list[Image] | None) -> Image | None:
    if not isinstance(image_list, list):
        return None
    covers = [image for image in image_list if image.type == "cover"]
    if len(covers) != 1:
        return None
    return covers[0]


def move_to_target_dir(source: str | Path, target: str | Path) -> str:
    return shutil.move(str(source), str(target))


def write_zip(zip_path: Path, entries: Iterable[tuple[str, ZipContent]]) -> None:
    with zipfile.ZipFile(
        zip_path, "w", compression=zipfile.ZIP_DEFLATED, allowZip64=True
    ) as zip_file:
        for arcname, content in entries:
            if isinstance(content, Path):
                zip_file.write(content, arcname=arcname)
            else:
                zip_file.writestr(arcname, content)



"""
Export recipes to Tandoor.

Tandoor expects a zip archive with a recipe.json file and an optional image.jpg.
"""

import json
import logging
import tempfile
from pathlib import Path
from typing import Any

import httpx

from kptncook.config import settings
from kptncook.exporter_utils import (
    asciify_string,
    get_cover,
    move_to_target_dir,
    write_zip,
    ZipContent,
)
from kptncook.ingredient_groups import iter_ingredient_groups
from kptncook.models import (
    Ingredient,
    localized_fallback,
    Recipe,
    RecipeStep,
    StepIngredient,
    StepIngredientUnit,
)

logger = logging.getLogger(__name__)


class TandoorExporter:
    def export(self, recipes: list[Recipe]) -> list[str]:
        filenames = []
        for recipe in recipes:
            filenames.append(self.export_recipe(recipe=recipe))
        return filenames

    def export_recipe(self, recipe: Recipe) -> str:
        filename = self.get_export_filename(recipe=recipe)
        payload = self.get_recipe_payload(recipe=recipe)
        image_bytes = self.get_cover_image_bytes(recipe=recipe)
        with tempfile.TemporaryDirectory() as tmp_dir:
            zip_path = Path(tmp_dir) / filename
            entries: list[tuple[str, ZipContent]] = [
                (
                    "recipe.json",
                    json.dumps(payload, ensure_ascii=False, indent=2),
                )
            ]
            if image_bytes is not None:
                entries.append(("image.jpg", image_bytes))
            write_zip(zip_path, entries)
            move_to_target_dir(zip_path, Path.cwd() / filename)
        return filename

    def get_export_filename(self, recipe: Recipe) -> str:
        title = localized_fallback(recipe.localized_title) or "kptncook-recipe"
        return f"{asciify_string(title)}.zip"

    def get_cover_image_bytes(self, recipe: Recipe) -> bytes | None:
        cover = get_cover(image_list=recipe.image_list)
        if cover is None:
            return None
        cover_url = recipe.get_image_url(api_key=settings.kptncook_api_key)
        if cover_url is None:
            return None
        try:
            response = httpx.get(cover_url, follow_redirects=True)
            response.raise_for_status()
        except httpx.HTTPStatusError as exc:
            if exc.response.status_code == 404:
                logger.info(
                    'Cover image for "%s" not found online any more.',
                    localized_fallback(recipe.localized_title) or "kptncook-recipe",
                )
            else:
                logger.warning(
                    "While trying to fetch the cover img a HTTP error occurred: %s: %s",
                    exc.response.status_code,
                    exc,
                )
            return None
        except httpx.RequestError as exc:
            logger.warning(
                "While trying to fetch the cover img a network error occurred: %s",
                exc,
            )
            return None
        return response.content

    def get_recipe_payload(self, recipe: Recipe) -> dict[str, Any]:
        return {
            "name": localized_fallback(recipe.localized_title) or "",
            "description": localized_fallback(recipe.author_comment) or "",
            "servings": 3,
            "source_url": self.get_source_url(recipe=recipe),
            "prep_time": recipe.preparation_time,
            "cook_time": recipe.cooking_time,
            "keywords": self.get_keywords(recipe=recipe),
            "steps": self.get_steps(recipe=recipe),
            "ingredients": self.get_ingredients(recipe=recipe),
        }

    def get_source_url(self, recipe: Recipe) -> str:
        source_id = recipe.uid or recipe.id.oid
        return f"https://share.kptncook.com/{source_id}"

    def get_keywords(self, recipe: Recipe) -> list[str]:
        keywords = ["kptncook"]
        if recipe.active_tags:
            keywords.extend(self._filter_active_tags(recipe.active_tags))
        if recipe.rtype:
            keywords.append(recipe.rtype)
        return keywords

    @staticmethod
    def _filter_active_tags(active_tags: list[str]) -> list[str]:
        return [tag for tag in active_tags if tag != "kptncook"]

    def get_steps(self, recipe: Recipe) -> list[dict[str, Any]]:
        steps = []
        for step in recipe.steps:
            steps.append(
                {
                    "instruction": localized_fallback(step.title) or "",
                    "ingredients": self.get_step_ingredients(step=step),
                }
            )
        return steps

    def get_step_ingredients(self, step: RecipeStep) -> list[dict[str, Any]]:
        ingredients = []
        for ingredient in step.ingredients or []:
            payload = self.get_step_ingredient_payload(ingredient=ingredient)
            if payload is not None:
                ingredients.append(payload)
        return ingredients

    def get_step_ingredient_payload(
        self, ingredient: StepIngredient | None
    ) -> dict[str, Any] | None:
        if ingredient is None:
            return None
        ingredient_name = self.get_step_ingredient_name(ingredient=ingredient) or ""
        payload = {
            "amount": ingredient.quantity,
            "unit": self.get_step_unit_payload(unit=ingredient.unit),
            "food": {"name": ingredient_name},
        }
        return {key: value for key, value in payload.items() if value is not None}

    def get_step_ingredient_name(self, ingredient: StepIngredient) -> str | None:
        details = ingredient.ingredient
        if details is None:
            return None
        if details.uncountable_title:
            title = localized_fallback(details.uncountable_title)
            if title:
                return title
        if details.localized_title:
            title = localized_fallback(details.localized_title)
            if title:
                return title
        if details.number_title:
            title = localized_fallback(details.number_title)
            if title:
                return title
        return None

    def get_step_unit_payload(
        self, unit: StepIngredientUnit | None
    ) -> dict[str, str] | None:
        if unit is None:
            return None
        for candidate in (
            unit.name,
            localized_fallback(unit.localized_title),
            localized_fallback(unit.short_title),
        ):
            if candidate:
                return {"name": candidate}
        return None

    def get_ingredients(self, recipe: Recipe) -> list[dict[str, Any]]:
        ingredients = []
        for group_label, group_ingredients in iter_ingredient_groups(
            recipe.ingredients
        ):
            if group_label:
                ingredients.append(self.get_group_header_payload(label=group_label))
            for ingredient in group_ingredients:
                ingredients.append(self.get_ingredient_payload(ingredient=ingredient))
        return ingredients

    def get_group_header_payload(self, label: str) -> dict[str, Any]:
        return {"note": label, "is_header": True, "no_amount": True, "amount": 0}

    def get_ingredient_payload(self, ingredient: Ingredient) -> dict[str, Any]:
        ingredient_name = self.get_ingredient_name(ingredient=ingredient) or ""
        payload = {
            "amount": ingredient.quantity,
            "unit": self.get_unit_payload(measure=ingredient.measure),
            "food": {"name": ingredient_name},
        }
        return {key: value for key, value in payload.items() if value is not None}

    def get_ingredient_name(self, ingredient: Ingredient) -> str | None:
        details = ingredient.ingredient
        if details.uncountable_title is not None:
            title = localized_fallback(details.uncountable_title)
            if title:
                return title
        return localized_fallback(details.localized_title)

    def get_unit_payload(self, measure: str | None) -> dict[str, str] | None:
        if measure is None:
            return None
        return {"name": measure}



"""
Setup entrypoint for first-run configuration without importing kptncook.
"""

# NOTE: This module intentionally duplicates env helpers and credential prompts
# from the main package to avoid importing kptncook.__init__, which triggers
# settings validation. This keeps setup runnable before any config exists.

from __future__ import annotations

import subprocess
import sys
from pathlib import Path
from urllib.parse import urljoin

import httpx
import typer
from rich import print as rprint
from rich.prompt import Prompt

DEFAULT_API_KEY = "6q7QNKy-oIgk-IMuWisJ-jfN7s6"
DEFAULT_API_URL = "https://mobile.kptncook.com"
ENV_PATH = Path.home() / ".kptncook" / ".env"
ENV_TEMPLATE = f"""# kptncook configuration
#
# Required
KPTNCOOK_API_KEY={DEFAULT_API_KEY}
#
# Optional: access token for favorites
KPTNCOOK_ACCESS_TOKEN=
#
# Optional: password manager integration
# KPTNCOOK_USERNAME_COMMAND="op read op://Personal/KptnCook/username"
# KPTNCOOK_PASSWORD_COMMAND="op read op://Personal/KptnCook/password"
"""

DEFAULT_HEADERS = {
    "content-type": "application/json",
    "Accept": "application/vnd.kptncook.mobile-v8+json",
    "User-Agent": "Platform/Android/12.0.1 App/7.10.1",
    "hasIngredients": "yes",
}

cli = typer.Typer()


def _scaffold_env_file(env_path: Path) -> bool:
    try:
        if env_path.exists() and env_path.stat().st_size > 0:
            return False
    except OSError:
        return False
    env_path.parent.mkdir(parents=True, exist_ok=True)
    try:
        env_path.write_text(ENV_TEMPLATE)
    except OSError:
        return False
    return True


def _read_env_values(env_path: Path) -> dict[str, str]:
    values: dict[str, str] = {}
    try:
        content = env_path.read_text()
    except OSError:
        return values
    for line in content.splitlines():
        stripped = line.strip()
        if not stripped or stripped.startswith("#") or "=" not in stripped:
            continue
        key, value = stripped.split("=", 1)
        values[key.strip()] = value.strip().strip('"')
    return values


def _upsert_env_value(env_path: Path, key: str, value: str) -> None:
    try:
        content = env_path.read_text()
        lines = content.splitlines()
    except OSError:
        lines = []
    updated = False
    new_lines: list[str] = []
    for line in lines:
        if line.strip().startswith(f"{key}="):
            new_lines.append(f"{key}={value}")
            updated = True
        else:
            new_lines.append(line)
    if not updated:
        if new_lines and new_lines[-1].strip() != "":
            new_lines.append("")
        new_lines.append(f"{key}={value}")
    env_path.parent.mkdir(parents=True, exist_ok=True)
    env_path.write_text("\n".join(new_lines) + "\n")


def _extract_http_error_message(response: httpx.Response) -> str | None:
    if "application/json" not in response.headers.get("content-type", ""):
        return None
    try:
        response_data = response.json()
    except ValueError:
        return None
    if isinstance(response_data, dict):
        for key in ("message", "error"):
            value = response_data.get(key)
            if isinstance(value, str):
                return value
        detail = response_data.get("detail")
        if isinstance(detail, dict):
            detail_message = detail.get("message")
            if isinstance(detail_message, str):
                return detail_message
    return None


def _fetch_access_token(api_key: str, username: str, password: str) -> str:
    headers = DEFAULT_HEADERS | {"kptnkey": api_key}
    response = httpx.post(
        urljoin(DEFAULT_API_URL, "/auth/login"),
        json={"email": username, "password": password},
        headers=headers,
    )
    response.raise_for_status()
    payload = response.json()
    return payload["accessToken"]


def _credential_from_command(command: str) -> str | None:
    try:
        result = subprocess.run(
            command, shell=True, capture_output=True, text=True, check=True
        )
        return result.stdout.strip()
    except subprocess.CalledProcessError as exc:
        rprint(f"[red]Error executing password manager command: {exc}[/red]")
        if exc.stderr:
            rprint(f"[red]Command output: {exc.stderr}[/red]")
        return None
    except Exception as exc:
        rprint(f"[red]Unexpected error: {exc}[/red]")
        return None


def _get_credentials(
    username_command: str | None, password_command: str | None
) -> tuple[str | None, str | None]:
    username = None
    password = None
    if username_command:
        username = _credential_from_command(username_command)
        if username:
            rprint("[green]✓ Username retrieved from password manager[/green]")
    if password_command:
        password = _credential_from_command(password_command)
        if password:
            rprint("[green]✓ Password retrieved from password manager[/green]")
    if not username:
        username = Prompt.ask("Enter your kptncook email address")
    if not password:
        password = Prompt.ask("Enter your kptncook password", password=True)
    return username, password


@cli.command()
def setup(
    env_path: Path = typer.Option(
        ENV_PATH, "--env-path", help="Path to the .env file to configure."
    ),
    api_key: str | None = typer.Option(
        None,
        "--api-key",
        help="KptnCook API key (defaults to the value documented in README.md).",
    ),
    username_command: str | None = typer.Option(
        None,
        "--username-command",
        help="Command to retrieve username from a password manager.",
    ),
    password_command: str | None = typer.Option(
        None,
        "--password-command",
        help="Command to retrieve password from a password manager.",
    ),
    fetch_access_token: bool = typer.Option(
        True,
        "--fetch-access-token/--no-fetch-access-token",
        help="Prompt for credentials and fetch an access token.",
    ),
) -> None:
    """
    Interactive setup for kptncook configuration.
    """
    scaffolded = _scaffold_env_file(env_path)
    if scaffolded:
        rprint(f"Created {env_path} with a starter template.")

    values = _read_env_values(env_path)
    existing_api_key = values.get("KPTNCOOK_API_KEY", "").strip()
    if not existing_api_key:
        chosen_key = api_key or DEFAULT_API_KEY
        _upsert_env_value(env_path, "KPTNCOOK_API_KEY", chosen_key)
        rprint("Added KPTNCOOK_API_KEY to the .env file.")
    else:
        chosen_key = existing_api_key
        rprint("KPTNCOOK_API_KEY is already set.")

    if username_command:
        _upsert_env_value(env_path, "KPTNCOOK_USERNAME_COMMAND", username_command)
        rprint("Stored KPTNCOOK_USERNAME_COMMAND in the .env file.")
    if password_command:
        _upsert_env_value(env_path, "KPTNCOOK_PASSWORD_COMMAND", password_command)
        rprint("Stored KPTNCOOK_PASSWORD_COMMAND in the .env file.")

    if not fetch_access_token:
        rprint("Skipping access token fetch.")
        return

    values = _read_env_values(env_path)
    username, password = _get_credentials(
        username_command or values.get("KPTNCOOK_USERNAME_COMMAND"),
        password_command or values.get("KPTNCOOK_PASSWORD_COMMAND"),
    )

    if not username or not password:
        rprint("[red]Failed to get credentials.[/red]")
        sys.exit(1)

    try:
        access_token = _fetch_access_token(chosen_key, username, password)
        _upsert_env_value(env_path, "KPTNCOOK_ACCESS_TOKEN", access_token)
        rprint("[green]✓ Access token retrieved successfully.[/green]")
        rprint(f"Saved KPTNCOOK_ACCESS_TOKEN to {env_path}.")
    except httpx.HTTPStatusError as exc:
        detail = _extract_http_error_message(exc.response)
        if exc.response.status_code == 401:
            message = (
                "Login failed (HTTP 401). Check your email/password and make sure "
                "KPTNCOOK_API_KEY is set to your real API key."
            )
        else:
            message = f"HTTP {exc.response.status_code} while getting access token"
            if detail:
                message = f"{message}: {detail}"
        rprint(f"[red]{message}[/red]")
        sys.exit(1)
    except httpx.HTTPError as exc:
        rprint(f"[red]Request failed: {exc}[/red]")
        sys.exit(1)



"""
Base settings for kptncook.
"""

import os
import sys
from pathlib import Path

from pydantic import AnyHttpUrl, DirectoryPath, Field, ValidationError, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict
from rich import print as rprint

from .env import ENV_PATH, scaffold_env_file


def _missing_required_fields(error: ValidationError) -> set[str]:
    missing: set[str] = set()
    for err in error.errors():
        loc = err.get("loc") or ()
        if not loc:
            continue
        field = str(loc[-1])
        err_type = err.get("type")
        if err_type == "missing":
            missing.add(field)
        elif field == "kptncook_api_key" and err_type == "value_error":
            missing.add(field)
    return missing


def _render_missing_settings_message(
    missing_fields: set[str],
    env_path: Path,
    scaffolded: bool,
) -> None:
    if "kptncook_api_key" not in missing_fields:
        return
    rprint("[red]Missing required configuration.[/red]")
    if scaffolded:
        rprint(f"Created {env_path} with a starter template.")
    rprint("Add your KptnCook API key to the .env file or your shell:")
    rprint(f"  {env_path}")
    rprint("  KPTNCOOK_API_KEY=your-api-key-here")
    rprint("See README.md for details.")


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_file=ENV_PATH, extra="ignore")
    root: DirectoryPath = Field(
        Path.home() / ".kptncook", validation_alias="KPTNCOOK_HOME"
    )  # type: ignore
    kptncook_api_key: str
    kptncook_access_token: str | None = None
    kptncook_api_url: AnyHttpUrl = AnyHttpUrl("https://mobile.kptncook.com")
    kptncook_lang: str = "de"
    kptncook_store: str = "de"
    kptncook_preferences: str | None = None
    mealie_url: AnyHttpUrl = AnyHttpUrl("http://localhost:9000/api")
    mealie_username: str | None = None
    mealie_password: str | None = None
    mealie_api_token: str | None = None

    # Password manager integration
    kptncook_username_command: str | None = None
    kptncook_password_command: str | None = None
    kptncook_group_ingredients_by_typ: bool = False
    kptncook_ingredient_group_labels: str | None = None

    @field_validator("root", mode="before")
    def root_must_exist(cls, path: str | Path | os.PathLike[str]) -> Path:
        path = Path(os.path.expandvars(os.fspath(path))).expanduser()
        path.mkdir(parents=True, exist_ok=True)
        return path

    @field_validator("kptncook_api_key")
    def api_key_must_be_set(cls, value: str) -> str:
        value = value.strip()
        if not value:
            raise ValueError("KPTNCOOK_API_KEY is required")
        return value


try:
    settings = Settings()  # type: ignore
except ValidationError as e:
    missing_fields = _missing_required_fields(e)
    scaffolded = False
    if "kptncook_api_key" in missing_fields:
        scaffolded = scaffold_env_file(ENV_PATH)
        _render_missing_settings_message(missing_fields, ENV_PATH, scaffolded)
    else:
        rprint("validation error: ", e)
    sys.exit(1)



from collections.abc import Iterable

from kptncook.config import settings
from kptncook.models import Ingredient

DEFAULT_INGREDIENT_GROUP_LABELS = {
    "regular": "You need",
    "basic": "Pantry",
}


def parse_ingredient_group_labels(raw: str | None) -> dict[str, str]:
    labels = dict(DEFAULT_INGREDIENT_GROUP_LABELS)
    if not raw:
        return labels
    for entry in raw.split(","):
        entry = entry.strip()
        if not entry or ":" not in entry:
            continue
        key, value = entry.split(":", 1)
        key = key.strip().lower()
        value = value.strip()
        if not key or not value:
            continue
        labels[key] = value
    return labels


def _ingredient_group_key(ingredient: Ingredient) -> str:
    details = getattr(ingredient, "ingredient", None)
    key = getattr(details, "typ", None)
    if isinstance(key, str) and key.strip():
        return key.strip().lower()
    return "other"


def _format_group_label(key: str, label_map: dict[str, str]) -> str:
    if key in label_map:
        return label_map[key]
    return key.replace("_", " ").title()


def iter_ingredient_groups(
    ingredients: Iterable[Ingredient],
) -> list[tuple[str | None, list[Ingredient]]]:
    items = list(ingredients)
    if not settings.kptncook_group_ingredients_by_typ:
        return [(None, items)]
    label_map = parse_ingredient_group_labels(settings.kptncook_ingredient_group_labels)
    groups: dict[str, list[Ingredient]] = {}
    seen_order: list[str] = []
    for ingredient in items:
        key = _ingredient_group_key(ingredient)
        if key not in groups:
            groups[key] = []
            seen_order.append(key)
        groups[key].append(ingredient)
    ordered_keys = [key for key in label_map if key in groups]
    for key in seen_order:
        if key not in ordered_keys:
            ordered_keys.append(key)
    return [(_format_group_label(key, label_map), groups[key]) for key in ordered_keys]



import re
from collections.abc import Sequence
from datetime import date
from time import time
from typing import Any, Literal
from urllib.parse import urljoin

import httpx

from .config import settings
from .repositories import RecipeInDb

RecipeIdentifier = tuple[Literal["oid", "uid"], str]


def ids_to_payload(ids: list[RecipeIdentifier]) -> list[dict]:
    """
    Convert a list of (type, id) tuples to a list of dicts that
    can be used as payload for the kptncook api.
    """
    payload = []
    for id_type, id_value in ids:
        if id_type == "oid":
            payload.append({"identifier": id_value})
        elif id_type == "uid":
            payload.append({"uid": id_value})
    return payload


def _coerce_recipe_identifier(value: object) -> RecipeIdentifier | None:
    if isinstance(value, tuple) and len(value) == 2:
        id_type, id_value = value
        if id_type in ("oid", "uid") and id_value is not None:
            return id_type, str(id_value)
    if isinstance(value, dict):
        identifier = value.get("identifier")
        if identifier is not None:
            nested = _coerce_recipe_identifier(identifier)
            if nested is not None:
                return nested if nested[0] == "oid" else ("oid", nested[1])
            if not isinstance(identifier, dict):
                return "oid", str(identifier)
        uid = value.get("uid")
        if uid is not None:
            nested = _coerce_recipe_identifier(uid)
            if nested is not None:
                return nested if nested[0] == "uid" else ("uid", nested[1])
            if not isinstance(uid, dict):
                return "uid", str(uid)
        for key in ("_id", "id", "recipeId", "recipe_id"):
            if key in value:
                nested = _coerce_recipe_identifier(value.get(key))
                if nested is not None:
                    return nested
        for key in ("$oid", "oid"):
            if key in value:
                nested = _coerce_recipe_identifier(value.get(key))
                if nested is not None:
                    return nested if nested[0] == "oid" else ("oid", nested[1])
        for key in ("recipe", "recipeSummary", "recipe_summary", "summary"):
            if key in value:
                nested = _coerce_recipe_identifier(value.get(key))
                if nested is not None:
                    return nested
    if isinstance(value, str):
        parsed = parse_id(value)
        if parsed is not None:
            return parsed
        if looks_like_uid(value):
            return "uid", value
        if looks_like_oid(value):
            return "oid", value
    return None


def _collect_recipe_identifiers(items: Sequence[object]) -> list[RecipeIdentifier]:
    identifiers: list[RecipeIdentifier] = []
    seen: set[RecipeIdentifier] = set()
    for item in items:
        parsed = _coerce_recipe_identifier(item)
        if parsed is None or parsed in seen:
            continue
        identifiers.append(parsed)
        seen.add(parsed)
    return identifiers


def _extract_dailies_payload(payload: object) -> list[dict[str, Any]]:
    if isinstance(payload, list):
        return [item for item in payload if isinstance(item, dict)]
    if isinstance(payload, dict):
        for key in ("recipes", "dailies", "items"):
            value = payload.get(key)
            if isinstance(value, list):
                return [item for item in value if isinstance(item, dict)]
    return []


def _extract_discovery_list_payload(payload: object) -> list[object]:
    if isinstance(payload, list):
        return payload
    if isinstance(payload, dict):
        for key in ("recipes", "items", "list", "data"):
            value = payload.get(key)
            if isinstance(value, list):
                return value
        for value in payload.values():
            if isinstance(value, dict):
                nested = _extract_discovery_list_payload(value)
                if nested:
                    return nested
    return []


_FAVORITES_KEYS = ("favorites", "items", "recipes", "data", "results")


def _extract_favorites_payload(
    payload: object,
) -> tuple[list[object], bool, bool]:
    if isinstance(payload, list):
        return payload, True, False
    if isinstance(payload, dict):
        for key in _FAVORITES_KEYS:
            if key in payload:
                value = payload.get(key)
                if isinstance(value, list):
                    return value, True, False
                if isinstance(value, dict):
                    nested, found, invalid = _extract_favorites_payload(value)
                    if found:
                        return nested, True, invalid
                return [], True, True
        for value in payload.values():
            if isinstance(value, dict):
                nested, found, invalid = _extract_favorites_payload(value)
                if found:
                    return nested, True, invalid
    return [], False, False


class KptnCookClient:
    """
    Client for the kptncook api.
    """

    def __init__(
        self, base_url=settings.kptncook_api_url, api_key=settings.kptncook_api_key
    ):
        self.base_url = str(base_url)
        self.headers = {
            "content-type": "application/json",
            "Accept": "application/vnd.kptncook.mobile-v8+json",
            "User-Agent": "Platform/Android/12.0.1 App/7.10.1",
            "hasIngredients": "yes",
        }
        self.api_key = api_key
        if settings.kptncook_access_token is not None:
            self.headers["Token"] = settings.kptncook_access_token

    @property
    def logged_in(self):
        return "Token" in self.headers

    def to_url(self, path):
        return urljoin(self.base_url, path)

    def __getattr__(self, name):
        """
        Return proxy for httpx, joining base_url with path and
        providing authentication headers automatically.
        """

        def proxy(path, **kwargs):
            url = self.to_url(path)
            set_headers = kwargs.get("headers", {})
            kwargs["headers"] = set_headers | self.headers
            return getattr(httpx, name)(url, **kwargs)

        return proxy

    def _standard_query_params(
        self,
        *,
        lang: str | None = None,
        store: str | None = None,
        preferences: str | None = None,
    ) -> dict[str, str]:
        if lang is None:
            lang = settings.kptncook_lang
        if store is None:
            store = settings.kptncook_store
        if preferences is None:
            preferences = settings.kptncook_preferences
        params = {"kptnkey": str(self.api_key)}
        if lang is not None:
            params["lang"] = str(lang)
        if store is not None:
            params["store"] = str(store)
        if preferences is not None:
            params["preferences"] = str(preferences)
        return params

    def list_today(self) -> list[RecipeInDb]:
        """
        Get all recipes for today from kptncook api.
        """
        time_str = str(time())
        response = self.get(f"recipes/de/{time_str}?kptnkey={self.api_key}")
        response.raise_for_status()
        recipes = []
        today = date.today()
        for data in response.json():
            recipes.append(RecipeInDb(date=today, data=data))
        return recipes

    def list_dailies(
        self,
        *,
        recipe_filter: str | None = None,
        zone: str | None = None,
        is_subscribed: bool | None = None,
        lang: str | None = None,
        store: str | None = None,
        preferences: str | None = None,
    ) -> list[RecipeInDb]:
        """
        Get daily recipes from kptncook api.
        """
        params = self._standard_query_params(
            lang=lang, store=store, preferences=preferences
        )
        if recipe_filter is not None:
            params["recipeFilter"] = recipe_filter
        if zone is not None:
            params["zone"] = zone
        if is_subscribed is not None:
            params["isSubscribed"] = "true" if is_subscribed else "false"
        response = self.get("/dailies", params=params)
        response.raise_for_status()
        payload = _extract_dailies_payload(response.json())
        today = date.today()
        return [RecipeInDb(date=today, data=data) for data in payload]

    def get_access_token(self, username: str, password: str) -> str:
        """
        Get access token for kptncook api.
        """
        headers = self.headers.copy()
        headers["kptnkey"] = str(self.api_key)
        response = self.post(
            "/auth/login",
            json={"email": username, "password": password},
            headers=headers,
        )
        response.raise_for_status()
        token_data = response.json()
        return token_data["accessToken"]

    def list_favorites(self) -> list[object]:
        """
        Get a list of favorite recipes.
        """
        params = self._standard_query_params()
        response = self.get("/favorites", params=params)
        response.raise_for_status()
        try:
            payload = response.json()
        except ValueError as exc:
            raise ValueError("Favorites response was not valid JSON") from exc
        favorites, found, invalid = _extract_favorites_payload(payload)
        if found and not invalid:
            return favorites
        if invalid:
            raise ValueError("Favorites response did not contain a list of favorites.")
        if isinstance(payload, dict):
            available_keys = ", ".join(sorted(payload.keys()))
            raise ValueError(
                "Favorites response missing favorites list. "
                f"Response keys: {available_keys}"
            )
        raise ValueError("Favorites response missing favorites list.")

    def get_by_ids(self, ids: list[RecipeIdentifier]) -> list[RecipeInDb]:
        """
        Get recipes from a list of ids.
        """
        return self.resolve_recipe_summaries(ids)

    def resolve_recipe_summaries(self, items: Sequence[object]) -> list[RecipeInDb]:
        """
        Resolve recipe summary payloads or identifiers into full recipes.
        """
        identifiers = _collect_recipe_identifiers(items)
        if not identifiers:
            return []
        payload = ids_to_payload(identifiers)
        if not payload:
            return []
        # timeout disabled because saving more than 999 favorites didn't work for @brotkrume
        response = self.post(
            f"/recipes/search?kptnkey={self.api_key}", json=payload, timeout=None
        )
        response.raise_for_status()
        results = response.json()
        if results is None:
            results = []
        return [RecipeInDb(date=date.today(), data=data) for data in results]

    def get_discovery_screen(
        self,
        *,
        lang: str | None = None,
        store: str | None = None,
        preferences: str | None = None,
        version: int = 2,
    ) -> dict[str, Any] | list[Any]:
        """
        Get discovery screen payload from kptncook api.
        """
        params = self._standard_query_params(
            lang=lang, store=store, preferences=preferences
        )
        params["v"] = str(version)
        response = self.get("discovery/screen", params=params)
        response.raise_for_status()
        return response.json()

    def get_discovery_list(
        self,
        *,
        list_type: str,
        list_id: str | None = None,
        lang: str | None = None,
        store: str | None = None,
        preferences: str | None = None,
    ) -> list[object]:
        """
        Get discovery list summary entries from kptncook api.
        """
        params = self._standard_query_params(
            lang=lang, store=store, preferences=preferences
        )
        if list_type in ("latest", "recommended"):
            path = f"discovery/list/{list_type}"
        elif list_type in ("curated", "automated"):
            if list_id is None:
                raise ValueError("list_id is required for curated or automated lists")
            path = f"discovery/list/{list_type}/{list_id}"
        else:
            path = f"discovery/list/{list_type}"
        response = self.get(path, params=params)
        response.raise_for_status()
        return _extract_discovery_list_payload(response.json())

    def get_onboarding_recipes(
        self,
        *,
        tags: list[str],
        lang: str | None = None,
        store: str | None = None,
        preferences: str | None = None,
    ) -> list[object]:
        """
        Get onboarding recipe summary entries from kptncook api.
        """
        if not tags:
            return []
        params = self._standard_query_params(
            lang=lang, store=store, preferences=preferences
        )
        response = self.post(
            "recipes/onboarding",
            params=params,
            json={"tags": tags},
        )
        response.raise_for_status()
        return _extract_discovery_list_payload(response.json())

    def list_popular_ingredients(
        self,
        *,
        lang: str | None = None,
        store: str | None = None,
    ) -> list[dict[str, Any]]:
        """
        Get popular ingredients from kptncook api.
        """
        if not self.logged_in:
            raise RuntimeError("Token required for /ingredients/popular")
        params = self._standard_query_params(lang=lang, store=store)
        response = self.get("ingredients/popular", params=params)
        response.raise_for_status()
        payload = response.json()
        if isinstance(payload, list):
            return [item for item in payload if isinstance(item, dict)]
        if isinstance(payload, dict):
            for key in ("ingredients", "items", "data"):
                value = payload.get(key)
                if isinstance(value, list):
                    return [item for item in value if isinstance(item, dict)]
        return []

    def get_recipes_with_ingredients(
        self,
        *,
        ingredient_ids: list[str],
        lang: str | None = None,
        store: str | None = None,
        preferences: str | None = None,
    ) -> list[object]:
        """
        Get recipe summaries that match ingredient ids.
        """
        if not ingredient_ids:
            return []
        params = self._standard_query_params(
            lang=lang, store=store, preferences=preferences
        )
        response = self.post(
            "recipes/withIngredients",
            params=params,
            json={"ingredientIds": ingredient_ids},
        )
        response.raise_for_status()
        return _extract_discovery_list_payload(response.json())


def looks_like_uid(token: str) -> bool:
    correct_len = len(token) == 8 or len(token) == 7
    is_alnum = token.isalnum()
    return correct_len and is_alnum


def looks_like_oid(token: str) -> bool:
    correct_len = len(token) == 24
    is_hex = token.isalnum()
    return correct_len and is_hex


def parse_id(text: str) -> tuple[Literal["oid", "uid"], str] | None:
    """
    Parse recipe id (uid or uid) from url/text.
    """
    try:
        uid = next(part for part in re.split(r"/|\?", text) if looks_like_uid(part))
        return "uid", uid
    except StopIteration:
        pass
    try:
        oid = next(part for part in re.split(r" |,|/", text) if looks_like_oid(part))
        return "oid", oid
    except StopIteration:
        pass
    return None



"""
Environment configuration helpers for kptncook.
"""

from __future__ import annotations

from pathlib import Path

DEFAULT_API_KEY = "6q7QNKy-oIgk-IMuWisJ-jfN7s6"
ENV_PATH = Path.home() / ".kptncook" / ".env"
ENV_TEMPLATE = f"""# kptncook configuration
#
# Required
KPTNCOOK_API_KEY={DEFAULT_API_KEY}
#
# Optional: access token for favorites
KPTNCOOK_ACCESS_TOKEN=
#
# Optional: password manager integration
# KPTNCOOK_USERNAME_COMMAND="op read op://Personal/KptnCook/username"
# KPTNCOOK_PASSWORD_COMMAND="op read op://Personal/KptnCook/password"
"""


def scaffold_env_file(env_path: Path = ENV_PATH) -> bool:
    try:
        if env_path.exists() and env_path.stat().st_size > 0:
            return False
    except OSError:
        return False
    env_path.parent.mkdir(parents=True, exist_ok=True)
    try:
        env_path.write_text(ENV_TEMPLATE)
    except OSError:
        return False
    return True


def read_env_values(env_path: Path) -> dict[str, str]:
    values: dict[str, str] = {}
    try:
        content = env_path.read_text()
    except OSError:
        return values
    for line in content.splitlines():
        stripped = line.strip()
        if not stripped or stripped.startswith("#") or "=" not in stripped:
            continue
        key, value = stripped.split("=", 1)
        values[key.strip()] = value.strip().strip('"')
    return values


def upsert_env_value(env_path: Path, key: str, value: str) -> None:
    try:
        content = env_path.read_text()
        lines = content.splitlines()
    except OSError:
        lines = []
    updated = False
    new_lines: list[str] = []
    for line in lines:
        if line.strip().startswith(f"{key}="):
            new_lines.append(f"{key}={value}")
            updated = True
        else:
            new_lines.append(line)
    if not updated:
        if new_lines and new_lines[-1].strip() != "":
            new_lines.append("")
        new_lines.append(f"{key}={value}")
    env_path.parent.mkdir(parents=True, exist_ok=True)
    env_path.write_text("\n".join(new_lines) + "\n")



"""
kptncook is a little command line utility to download
new recipes.
"""

import sys
from datetime import date
from typing import Any, Optional

import httpx
import typer
from rich import print as rprint
from rich.pretty import pprint

from .api import KptnCookClient, _collect_recipe_identifiers, parse_id
from .config import settings
from .env import ENV_PATH
from .mealie import MealieApiClient, kptncook_to_mealie
from .models import Recipe, localized_fallback
from .paprika import PaprikaExporter
from .tandoor import TandoorExporter
from .repositories import RecipeInDb, RecipeRepository

__all__ = [
    "list_kptncook_today",
    "list_kptncook_dailies",
    "save_todays_recipes",
    "sync_with_mealie",
    "sync",
    "backup_kptncook_favorites",
    "get_kptncook_access_token",
    "list_recipes",
    "delete_recipes",
    "search_kptncook_recipe_by_id",
    "list_discovery_screen",
    "list_discovery_list",
    "list_onboarding_recipes",
    "list_popular_ingredients",
    "list_recipes_with_ingredients",
    "export_recipes_to_paprika",
    "export_recipes_to_tandoor",
]

__version__ = "0.0.28"
cli = typer.Typer()


@cli.command(name="kptncook-today")
def list_kptncook_today():
    """
    List all recipes for today from the kptncook site.
    """
    client = KptnCookClient()
    all_recipes = client.list_today()
    for recipe in all_recipes:
        pprint(recipe)


@cli.command(name="save-todays-recipes")
def save_todays_recipes():
    """
    Save recipes for today from kptncook site.
    """
    fs_repo = RecipeRepository(settings.root)
    if fs_repo.needs_to_be_synced(date.today()):
        client = KptnCookClient()
        fs_repo.add_list(client.list_today())


@cli.command(name="dailies")
def list_kptncook_dailies(
    recipe_filter: str | None = typer.Option(
        None, "--recipe-filter", help="Filter daily recipes by recipeFilter value."
    ),
    zone: str | None = typer.Option(
        None, "--zone", help="Filter daily recipes by zone."
    ),
    is_subscribed: bool | None = typer.Option(
        None,
        "--subscribed/--not-subscribed",
        help="Filter daily recipes by subscription status.",
    ),
    save: bool = typer.Option(
        False,
        "--save",
        "-s",
        help="Save daily recipes to the local repository.",
    ),
):
    """
    List daily recipes from the kptncook site.
    """
    client = KptnCookClient()
    try:
        recipes = client.list_dailies(
            recipe_filter=recipe_filter,
            zone=zone,
            is_subscribed=is_subscribed,
        )
    except httpx.HTTPStatusError as exc:
        detail = _extract_http_error_message(exc.response)
        message = f"HTTP {exc.response.status_code} while fetching dailies"
        if detail:
            message = f"{message}: {detail}"
        rprint(f"[red]{message}[/red]")
        sys.exit(1)
    except httpx.HTTPError as exc:
        rprint(f"[red]Request failed: {exc}[/red]")
        sys.exit(1)
    if not recipes:
        rprint("No recipes found.")
        return
    for recipe in recipes:
        pprint(recipe)
    if save:
        fs_repo = RecipeRepository(settings.root)
        fs_repo.add_list(recipes)
        rprint(f"Added {len(recipes)} recipes to local repository")


def get_mealie_client() -> MealieApiClient:
    client = MealieApiClient(settings.mealie_url)
    if settings.mealie_api_token:
        client.login_with_token(settings.mealie_api_token)
        return client
    if settings.mealie_username and settings.mealie_password:
        client.login(settings.mealie_username, settings.mealie_password)
        return client
    rprint(
        "[red]Mealie authentication required. "
        "Set MEALIE_API_TOKEN or MEALIE_USERNAME/MEALIE_PASSWORD.[/red]"
    )
    sys.exit(1)


def get_kptncook_recipes_from_mealie(client):
    recipes = client.get_all_recipes()
    recipes_with_details = []
    for recipe in recipes:
        recipes_with_details.append(client.get_via_slug(recipe.slug))
    kptncook_recipes = [
        r for r in recipes_with_details if r.extras.get("source") == "kptncook"
    ]
    return kptncook_recipes


def get_kptncook_recipes_from_repository() -> list[Recipe]:
    fs_repo = RecipeRepository(settings.root)
    recipes = []
    for repo_recipe in fs_repo.list():
        # recipes.append(Recipe.model_validate(repo_recipe.data))
        try:
            recipes.append(Recipe.model_validate(repo_recipe.data))
        except Exception:
            continue
    return recipes


def get_recipe_from_repository_by_oid(oid: str) -> list[Recipe]:
    """
    get one single recipe from local repository
    :param oid: oid of recipe
    :return: list
    """
    recipes = get_kptncook_recipes_from_repository()
    return [recipe for num, recipe in enumerate(recipes) if recipe.id.oid == oid]


def _extract_mealie_detail_message(response: httpx.Response) -> str | None:
    if "application/json" not in response.headers.get("content-type", ""):
        return None
    try:
        response_data = response.json()
    except ValueError:
        return None
    if isinstance(response_data, dict):
        detail = response_data.get("detail", {})
        if isinstance(detail, dict):
            detail_message = detail.get("message")
            if isinstance(detail_message, str):
                return detail_message
    return None


def _extract_http_error_message(response: httpx.Response) -> str | None:
    if "application/json" not in response.headers.get("content-type", ""):
        return None
    try:
        response_data = response.json()
    except ValueError:
        return None
    if isinstance(response_data, dict):
        for key in ("message", "error"):
            value = response_data.get(key)
            if isinstance(value, str):
                return value
        detail = response_data.get("detail")
        if isinstance(detail, dict):
            detail_message = detail.get("message")
            if isinstance(detail_message, str):
                return detail_message
    return None


def _resolve_recipe_summaries(
    client: KptnCookClient, items: list[object]
) -> list[RecipeInDb]:
    if not items:
        return []
    try:
        return client.resolve_recipe_summaries(items)
    except httpx.HTTPStatusError as exc:
        detail = _extract_http_error_message(exc.response)
        message = f"HTTP {exc.response.status_code} while resolving recipes"
        if detail:
            message = f"{message}: {detail}"
        rprint(f"[red]{message}[/red]")
        sys.exit(1)
    except httpx.HTTPError as exc:
        rprint(f"[red]Request failed: {exc}[/red]")
        sys.exit(1)


def _extract_nested_list(payload: object, keys: tuple[str, ...]) -> list[Any]:
    if isinstance(payload, dict):
        for key in keys:
            value = payload.get(key)
            if isinstance(value, list):
                return value
        for value in payload.values():
            if isinstance(value, dict):
                nested = _extract_nested_list(value, keys)
                if nested:
                    return nested
    return []


def _extract_localized_text(value: object) -> str | None:
    if isinstance(value, str):
        return value
    if isinstance(value, dict):
        for key in ("de", "en", "es", "fr", "pt"):
            candidate = value.get(key)
            if isinstance(candidate, str):
                return candidate
        for key in ("singular", "plural", "uncountable"):
            candidate = value.get(key)
            if isinstance(candidate, str):
                return candidate
        for candidate in value.values():
            if isinstance(candidate, str):
                return candidate
    return None


def _coerce_discovery_list_id(value: object) -> str | None:
    if isinstance(value, dict):
        if "$oid" in value:
            value = value["$oid"]
        elif "oid" in value:
            value = value["oid"]
    if isinstance(value, str):
        return value
    if value is None:
        return None
    return str(value)


def _extract_discovery_list_entries(payload: object) -> list[dict[str, Any]]:
    candidates = _extract_nested_list(
        payload, ("lists", "discoveryLists", "discoveryList", "items", "sections")
    )
    return [item for item in candidates if isinstance(item, dict)]


def _extract_discovery_list_id(entry: dict[str, Any]) -> str | None:
    for key in ("id", "_id", "listId", "list_id", "oid"):
        if key in entry:
            return _coerce_discovery_list_id(entry.get(key))
    return None


def _coerce_ingredient_id(value: object) -> str | None:
    if isinstance(value, dict):
        if "$oid" in value:
            value = value["$oid"]
        elif "oid" in value:
            value = value["oid"]
    if isinstance(value, str):
        return value
    if value is None:
        return None
    return str(value)


def _extract_ingredient_id(entry: dict[str, Any]) -> str | None:
    for key in ("_id", "id", "ingredientId", "ingredient_id", "oid"):
        if key in entry:
            return _coerce_ingredient_id(entry.get(key))
    return None


def _extract_ingredient_name(entry: dict[str, Any]) -> str | None:
    for key in ("numberTitle", "localizedTitle", "name", "title", "label"):
        if key in entry:
            name = _extract_localized_text(entry.get(key))
            if name:
                return name
    return None


def _extract_discovery_list_title(entry: dict[str, Any]) -> str | None:
    for key in ("title", "localizedTitle", "name"):
        if key in entry:
            title = _extract_localized_text(entry.get(key))
            if title:
                return title
    return None


def _extract_discovery_list_type(entry: dict[str, Any]) -> str | None:
    for key in ("listType", "list_type", "type"):
        value = entry.get(key)
        if isinstance(value, str):
            return value
    return None


def _extract_quick_search_entries(payload: object) -> list[Any]:
    return _extract_nested_list(
        payload,
        (
            "quickSearchEntries",
            "quickSearch",
            "quick_search",
            "quicksearch",
        ),
    )


def _format_quick_search_entry(entry: object) -> str | None:
    if isinstance(entry, str):
        return entry
    if isinstance(entry, dict):
        for key in ("title", "name", "label", "text", "query"):
            value = entry.get(key)
            if isinstance(value, str):
                return value
        localized = entry.get("localizedTitle") or entry.get("title")
        return _extract_localized_text(localized)
    return None


_DISCOVERY_LIST_TYPES = {"latest", "recommended", "curated", "automated"}
_DISCOVERY_LIST_TYPES_REQUIRE_ID = {"curated", "automated"}


def _normalize_discovery_list_type(value: str) -> str:
    normalized = value.strip().lower()
    if normalized not in _DISCOVERY_LIST_TYPES:
        raise typer.BadParameter(
            "list-type must be one of: latest, recommended, curated, automated"
        )
    return normalized


def _normalize_discovery_list_id(value: str | None) -> str | None:
    if value is None:
        return None
    normalized = value.strip()
    return normalized or None


def _normalize_tags(tags: list[str]) -> list[str]:
    return _normalize_list_input(tags)


def _normalize_list_input(items: list[str]) -> list[str]:
    normalized: list[str] = []
    seen: set[str] = set()
    for item in items:
        for raw in item.split(","):
            candidate = raw.strip()
            if not candidate or candidate in seen:
                continue
            normalized.append(candidate)
            seen.add(candidate)
    return normalized


def _normalize_ingredient_ids(ingredient_ids: list[str]) -> list[str]:
    return _normalize_list_input(ingredient_ids)


def _require_access_token() -> None:
    if settings.kptncook_access_token is None:
        rprint(f"Please set KPTNCOOK_ACCESS_TOKEN in your environment or {ENV_PATH}")
        sys.exit(1)


@cli.command(name="sync-with-mealie")
def sync_with_mealie():
    """
    Sync locally saved recipes with mealie.
    """
    try:
        client = get_mealie_client()
    except Exception as e:
        rprint(f"Could not login to mealie: {e}")
        sys.exit(1)
    kptncook_recipes_from_mealie = get_kptncook_recipes_from_mealie(client)
    recipes = get_kptncook_recipes_from_repository()
    kptncook_recipes_from_repository = [kptncook_to_mealie(r) for r in recipes]
    ids_in_mealie = {r.extras["kptncook_id"] for r in kptncook_recipes_from_mealie}
    ids_from_api = {r.extras["kptncook_id"] for r in kptncook_recipes_from_repository}
    ids_to_add = ids_from_api - ids_in_mealie
    recipes_to_add = []
    for recipe in kptncook_recipes_from_repository:
        if recipe.extras.get("kptncook_id") in ids_to_add:
            recipes_to_add.append(recipe)
    created_slugs = []
    for recipe in recipes_to_add:
        try:
            created = client.create_recipe(recipe)
            created_slugs.append(created.slug)
        except httpx.HTTPStatusError as e:
            detail_message = _extract_mealie_detail_message(e.response)
            if detail_message == "Recipe already exists":
                continue
    rprint(f"Created {len(created_slugs)} recipes")


@cli.command(name="sync")
def sync():
    """
    Fetch recipes for today from api, save them to disk and sync with mealie
    afterwards.
    """
    save_todays_recipes()
    sync_with_mealie()


@cli.command(name="backup-favorites")
def backup_kptncook_favorites():
    """
    Store kptncook favorites in local repository.
    """
    _require_access_token()
    client = KptnCookClient()
    try:
        favorites = client.list_favorites()
    except ValueError as exc:
        rprint(f"[red]{exc}[/red]")
        sys.exit(1)
    rprint(f"Found {len(favorites)} favorites")
    identifiers = _collect_recipe_identifiers(favorites)
    if not identifiers:
        rprint("Could not find any favorites")
        sys.exit(1)
    recipes = client.get_by_ids(identifiers)
    if len(recipes) == 0:
        rprint("Could not find any favorites")
        sys.exit(1)

    fs_repo = RecipeRepository(settings.root)
    fs_repo.add_list(recipes)
    rprint(f"Added {len(recipes)} recipes to local repository")


@cli.command(name="kptncook-access-token")
def get_kptncook_access_token():
    """
    Get access token for kptncook.

    Credentials can be retrieved from a password manager by setting:
    - KPTNCOOK_USERNAME_COMMAND: Command to retrieve username
    - KPTNCOOK_PASSWORD_COMMAND: Command to retrieve password

    Example for 1Password:
    KPTNCOOK_USERNAME_COMMAND="op read op://Personal/KptnCook/username"
    KPTNCOOK_PASSWORD_COMMAND="op read op://Personal/KptnCook/password"
    """
    from .password_manager import get_credentials

    username, password = get_credentials(
        username_command=settings.kptncook_username_command,
        password_command=settings.kptncook_password_command,
    )

    if not username or not password:
        rprint("[red]Failed to get credentials[/red]")
        sys.exit(1)

    client = KptnCookClient()
    try:
        access_token = client.get_access_token(username, password)
        rprint("[green]✓ Access token retrieved successfully[/green]")
        rprint("Your access token: ", access_token)
        rprint(f"Add it to {ENV_PATH} as KPTNCOOK_ACCESS_TOKEN=...")
    except httpx.HTTPStatusError as exc:
        detail = _extract_http_error_message(exc.response)
        if exc.response.status_code == 401:
            message = (
                "Login failed (HTTP 401). Check your email/password and make sure "
                "KPTNCOOK_API_KEY is set to your real API key (not a placeholder)."
            )
        else:
            message = f"HTTP {exc.response.status_code} while getting access token"
            if detail:
                message = f"{message}: {detail}"
        rprint(f"[red]{message}[/red]")
        sys.exit(1)
    except httpx.HTTPError as exc:
        rprint(f"[red]Request failed: {exc}[/red]")
        sys.exit(1)


@cli.command(name="list-recipes")
def list_recipes():
    """
    List all locally saved recipes.
    """
    recipes = get_kptncook_recipes_from_repository()
    for num, recipe in enumerate(recipes):
        title = localized_fallback(recipe.localized_title) or "Unknown title"
        rprint(num, title, recipe.id.oid)


@cli.command(name="discovery-screen")
def list_discovery_screen(
    show_quick_search: bool = typer.Option(
        True, "--quick-search/--no-quick-search", help="Show quick search entries."
    ),
):
    """
    List discovery screen lists and quick search entries.
    """
    client = KptnCookClient()
    try:
        payload = client.get_discovery_screen()
    except httpx.HTTPStatusError as exc:
        detail = _extract_http_error_message(exc.response)
        message = f"HTTP {exc.response.status_code} while fetching discovery screen"
        if detail:
            message = f"{message}: {detail}"
        rprint(f"[red]{message}[/red]")
        sys.exit(1)
    except httpx.HTTPError as exc:
        rprint(f"[red]Request failed: {exc}[/red]")
        sys.exit(1)

    list_entries = _extract_discovery_list_entries(payload)
    list_printed = 0
    for entry in list_entries:
        list_id = _extract_discovery_list_id(entry)
        title = _extract_discovery_list_title(entry)
        list_type = _extract_discovery_list_type(entry)
        if list_id is None and title is None and list_type is None:
            continue
        if list_printed == 0:
            rprint("Discovery lists:")
        list_printed += 1
        list_id = list_id or "-"
        title = title or "-"
        list_type = list_type or "-"
        rprint(f"- {list_id} | {title} | {list_type}")
    if list_printed == 0:
        rprint("No discovery lists found.")

    if show_quick_search:
        quick_search = _extract_quick_search_entries(payload)
        quick_printed = 0
        for entry in quick_search:
            label = _format_quick_search_entry(entry)
            if label is None:
                continue
            if quick_printed == 0:
                rprint("Quick search:")
            quick_printed += 1
            rprint(f"- {label}")
        if quick_printed == 0:
            rprint("No quick search entries found.")


@cli.command(name="discovery-list")
def list_discovery_list(
    list_type: str = typer.Option(
        ...,
        "--list-type",
        "-t",
        help="Discovery list type (latest, recommended, curated, automated).",
    ),
    list_id: str | None = typer.Option(
        None,
        "--list-id",
        "-i",
        help="Discovery list id (required for curated/automated).",
    ),
    save: bool = typer.Option(
        False,
        "--save",
        "-s",
        help="Save discovery list recipes to the local repository.",
    ),
):
    """
    List recipes from a discovery list.
    """
    list_type = _normalize_discovery_list_type(list_type)
    list_id = _normalize_discovery_list_id(list_id)
    if list_type in _DISCOVERY_LIST_TYPES_REQUIRE_ID and list_id is None:
        raise typer.BadParameter(
            "list-id is required when list-type is curated or automated"
        )

    client = KptnCookClient()
    try:
        items = client.get_discovery_list(list_type=list_type, list_id=list_id)
    except httpx.HTTPStatusError as exc:
        detail = _extract_http_error_message(exc.response)
        message = f"HTTP {exc.response.status_code} while fetching discovery list"
        if detail:
            message = f"{message}: {detail}"
        rprint(f"[red]{message}[/red]")
        sys.exit(1)
    except httpx.HTTPError as exc:
        rprint(f"[red]Request failed: {exc}[/red]")
        sys.exit(1)

    if not items:
        rprint("No discovery list recipes found.")
        return

    recipes = _resolve_recipe_summaries(client, items)
    if not recipes:
        rprint("No recipes found.")
        return

    for recipe in recipes:
        pprint(recipe)

    if save:
        fs_repo = RecipeRepository(settings.root)
        fs_repo.add_list(recipes)
        rprint(f"Added {len(recipes)} recipes to local repository")


@cli.command(name="ingredients-popular")
def list_popular_ingredients():
    """
    List popular ingredients.
    """
    _require_access_token()
    client = KptnCookClient()
    try:
        ingredients = client.list_popular_ingredients()
    except httpx.HTTPStatusError as exc:
        detail = _extract_http_error_message(exc.response)
        message = f"HTTP {exc.response.status_code} while fetching popular ingredients"
        if detail:
            message = f"{message}: {detail}"
        rprint(f"[red]{message}[/red]")
        sys.exit(1)
    except httpx.HTTPError as exc:
        rprint(f"[red]Request failed: {exc}[/red]")
        sys.exit(1)

    if not ingredients:
        rprint("No popular ingredients found.")
        return

    printed = 0
    for entry in ingredients:
        ingredient_id = _extract_ingredient_id(entry)
        name = _extract_ingredient_name(entry)
        if ingredient_id is None and name is None:
            continue
        if printed == 0:
            rprint("Popular ingredients:")
        printed += 1
        ingredient_id = ingredient_id or "-"
        name = name or "-"
        rprint(f"- {ingredient_id} | {name}")

    if printed == 0:
        rprint("No popular ingredients found.")


@cli.command(name="recipes-with-ingredients")
def list_recipes_with_ingredients(
    ingredient_ids: list[str] = typer.Option(
        ...,
        "--ingredient-id",
        "-i",
        help="Ingredient id (repeatable, comma-separated ok).",
    ),
    save: bool = typer.Option(
        False,
        "--save",
        "-s",
        help="Save recipes to the local repository.",
    ),
):
    """
    List recipes that match ingredient ids.
    """
    _require_access_token()
    ids = _normalize_ingredient_ids(ingredient_ids)
    if not ids:
        rprint("Please provide one or more non-empty --ingredient-id values.")
        sys.exit(1)

    client = KptnCookClient()
    try:
        items = client.get_recipes_with_ingredients(ingredient_ids=ids)
    except httpx.HTTPStatusError as exc:
        detail = _extract_http_error_message(exc.response)
        message = (
            f"HTTP {exc.response.status_code} while fetching recipes with ingredients"
        )
        if detail:
            message = f"{message}: {detail}"
        rprint(f"[red]{message}[/red]")
        sys.exit(1)
    except httpx.HTTPError as exc:
        rprint(f"[red]Request failed: {exc}[/red]")
        sys.exit(1)

    if not items:
        rprint("No recipes found.")
        return

    recipes = _resolve_recipe_summaries(client, items)
    if not recipes:
        rprint("No recipes found.")
        return

    for recipe in recipes:
        pprint(recipe)

    if save:
        fs_repo = RecipeRepository(settings.root)
        fs_repo.add_list(recipes)
        rprint(f"Added {len(recipes)} recipes to local repository")


@cli.command(name="onboarding")
def list_onboarding_recipes(
    tags: list[str] = typer.Option(
        ..., "--tag", "-t", help="Onboarding tag (repeatable, comma-separated ok)."
    ),
    save: bool = typer.Option(
        False,
        "--save",
        "-s",
        help="Save onboarding recipes to the local repository.",
    ),
):
    """
    List onboarding recipes by tags.
    """
    tag_list = _normalize_tags(tags)
    if not tag_list:
        rprint("Please provide one or more non-empty --tag values.")
        sys.exit(1)

    client = KptnCookClient()
    try:
        items = client.get_onboarding_recipes(tags=tag_list)
    except httpx.HTTPStatusError as exc:
        detail = _extract_http_error_message(exc.response)
        message = f"HTTP {exc.response.status_code} while fetching onboarding recipes"
        if detail:
            message = f"{message}: {detail}"
        rprint(f"[red]{message}[/red]")
        sys.exit(1)
    except httpx.HTTPError as exc:
        rprint(f"[red]Request failed: {exc}[/red]")
        sys.exit(1)

    if not items:
        rprint("No onboarding recipes found.")
        return

    recipes = _resolve_recipe_summaries(client, items)
    if not recipes:
        rprint("No recipes found.")
        return

    for recipe in recipes:
        pprint(recipe)

    if save:
        fs_repo = RecipeRepository(settings.root)
        fs_repo.add_list(recipes)
        rprint(f"Added {len(recipes)} recipes to local repository")


@cli.command(name="delete-recipes")
def delete_recipes(
    indices: list[int] | None = typer.Argument(
        None, help="Indices from list-recipes to delete."
    ),
    oids: list[str] | None = typer.Option(
        None, "--oid", "-o", help="Recipe oid to delete (repeatable)."
    ),
    force: bool = typer.Option(False, "--force", "-f", help="Skip confirmation."),
):
    """
    Delete recipes from the local repository.
    """
    index_list = indices or []
    oid_list = oids or []
    if not index_list and not oid_list:
        rprint("Please provide one or more recipe indices or --oid values.")
        sys.exit(1)

    recipes = get_kptncook_recipes_from_repository()
    index_ids = []
    invalid_indices = []
    for index in index_list:
        if index < 0 or index >= len(recipes):
            invalid_indices.append(index)
            continue
        index_ids.append(recipes[index].id.oid)

    requested_ids = []
    for oid in index_ids + oid_list:
        if oid not in requested_ids:
            requested_ids.append(oid)

    repo = RecipeRepository(settings.root)
    existing_by_id = repo.list_by_id()
    existing_ids = {str(key) for key in existing_by_id.keys()}

    missing_ids = [oid for oid in requested_ids if str(oid) not in existing_ids]
    to_delete_ids = [oid for oid in requested_ids if str(oid) in existing_ids]

    if invalid_indices:
        rprint(
            "Invalid indices (out of range): "
            + ", ".join(str(i) for i in invalid_indices)
        )
    if missing_ids:
        rprint("Unknown recipe ids: " + ", ".join(missing_ids))

    if not to_delete_ids:
        rprint("No matching recipes to delete.")
        sys.exit(1)

    recipe_by_oid = {recipe.id.oid: recipe for recipe in recipes}
    rprint("Recipes to delete:")
    for oid in to_delete_ids:
        recipe = recipe_by_oid.get(oid)
        if recipe is None:
            rprint(f"- {oid}")
            continue
        title = localized_fallback(recipe.localized_title) or "Unknown title"
        rprint(f"- {title} ({oid})")

    if not force and not typer.confirm("Delete these recipes from local storage?"):
        rprint("Aborted.")
        sys.exit(1)

    deleted, missing = repo.delete_by_ids(to_delete_ids)
    if missing:
        rprint("Some recipes were not found: " + ", ".join(missing))
    rprint(f"Deleted {len(deleted)} recipes.")


@cli.command(name="search-by-id")
def search_kptncook_recipe_by_id(id_: str):
    """
    Search for a recipe by id in kptncook api, id can be a sharing
    url or an oid for example, and add it to the local repository.
    """
    if id_.startswith(
        "https://share.kptncook.com/"
    ):  # sharing url -> use redirect location
        r = httpx.get(id_)
        if r.status_code not in (301, 302):
            rprint("Could not get redirect location")
            sys.exit(1)
        id_ = r.headers["location"]
    parsed = parse_id(id_)
    if parsed is None:
        rprint("Could not parse id")
        sys.exit(1)
    id_type, id_value = parsed
    rprint(id_type, id_value)
    client = KptnCookClient()
    recipes = client.get_by_ids([(id_type, id_value)])
    if len(recipes) == 0:
        rprint("Could not find recipe")
        sys.exit(1)
    recipe = recipes[0]
    fs_repo = RecipeRepository(settings.root)
    fs_repo.add_list([recipe])
    rprint(f"Added recipe {id_type} {id_value} to local repository")


# Optional needed by typer, standalone to trick pyupgrade to not change it
OptionalId = Optional[str]


@cli.command(name="export-recipes-to-paprika")
def export_recipes_to_paprika(_id: OptionalId = typer.Argument(None)):
    """
    Export one recipe or all recipes to Paprika app

    Example usage 1:  kptncook  export-recipes-to-paprika 635a68635100007500061cd7
    Example usage 2:  kptncook  export-recipes-to-paprika
    """
    if _id:
        recipes = get_recipe_by_id(_id)
    else:
        recipes = get_kptncook_recipes_from_repository()
    exporter = PaprikaExporter()
    filename = exporter.export(recipes=recipes)
    rprint(
        "\n The data was exported to '%s'. Open the export file with the Paprika App.\n"
        % filename
    )


@cli.command(name="export-recipes-to-tandoor")
def export_recipes_to_tandoor(_id: OptionalId = typer.Argument(None)):
    """
    Export one recipe or all recipes to Tandoor

    Example usage 1:  kptncook  export-recipes-to-tandoor 635a68635100007500061cd7
    Example usage 2:  kptncook  export-recipes-to-tandoor
    """
    if _id:
        recipes = get_recipe_by_id(_id)
    else:
        recipes = get_kptncook_recipes_from_repository()
    exporter = TandoorExporter()
    filenames = exporter.export(recipes=recipes)
    rprint(
        "\n The data was exported to '%s'. Open the export file with Tandoor.\n"
        % ", ".join(filenames)
    )


def get_recipe_by_id(_id: str):
    parsed = parse_id(_id)
    if parsed is None:
        rprint("Could not parse id")
        sys.exit(1)
    # we can expect always an oid here - correct?
    id_type, id_value = parsed
    found_recipes = get_recipe_from_repository_by_oid(oid=id_value)
    if len(found_recipes) == 0:
        rprint("Recipe not found.")
        sys.exit(1)
    if len(found_recipes) > 1:
        rprint("More than one recipe found with that ID.")
        sys.exit(1)
    return found_recipes


if __name__ == "__main__":
    cli()



"""
All the domain models for kptncook live here.
"""

from typing import Any

from pydantic import BaseModel, ConfigDict, Field, model_validator


def to_camel(string: str) -> str:
    """Convert snake_case to camelCase."""
    words = []
    for num, word in enumerate(string.split("_")):
        if num > 0:
            word = word.capitalize()
        words.append(word)
    return "".join(words)


class LocalizedString(BaseModel):
    en: str | None = None
    de: str | None = None
    es: str | None = None
    fr: str | None = None
    pt: str | None = None

    def fallback(self) -> str | None:
        for candidate in (self.de, self.en, self.es, self.fr, self.pt):
            if candidate:
                return candidate
        return None

    @model_validator(mode="before")
    def coerce_string(cls, value: Any) -> dict[str, Any] | Any:
        if isinstance(value, str):
            # Default to German for string-only payloads.
            return {"de": value}
        if isinstance(value, dict):
            singular = value.get("singular")
            plural = value.get("plural")
            uncountable = value.get("uncountable")
            for candidate in (singular, plural, uncountable):
                if isinstance(candidate, dict):
                    # Prefer localized singular/plural if present.
                    return candidate
            for candidate in (uncountable, singular, plural):
                if isinstance(candidate, str):
                    # Fall back to the raw string if only singular/plural are present.
                    return {"de": candidate}
        return value


class LocalizedStepIngredientUnit(LocalizedString):
    """Localized unit labels for step ingredients."""


def localized_fallback(value: LocalizedString | None) -> str | None:
    if value is None:
        return None
    return value.fallback()


class Nutrition(BaseModel):
    calories: int
    protein: int
    fat: int
    carbohydrate: int


class Image(BaseModel):
    name: str
    type: str | None = None
    url: str

    def get_image_with_api_key_url(self, api_key: str) -> "Image":
        url_with_key = f"{self.url}?kptnkey={api_key}"
        kwargs = self.model_dump() | {"url": url_with_key}
        return Image(**kwargs)


class RecipeId(BaseModel):
    oid: str = Field(..., alias="$oid")


class IngredientDetails(BaseModel):
    model_config = ConfigDict(alias_generator=to_camel)

    id: RecipeId | None = Field(None, alias="_id")
    typ: str
    localized_title: LocalizedString
    number_title: LocalizedString
    uncountable_title: LocalizedString | None = None
    category: str

    @model_validator(mode="before")
    def fix_json_errors(cls, values):
        if isinstance(values, dict):
            if values.get("localizedTitle") is None:
                for key in ("uncountableTitle", "title", "numberTitle", "name"):
                    if values.get(key) is not None:
                        values["localizedTitle"] = values[key]
                        break
            if values.get("numberTitle") is None:
                for key in ("localizedTitle", "uncountableTitle", "title", "name"):
                    if values.get(key) is not None:
                        values["numberTitle"] = values[key]
                        break
            if values.get("uncountableTitle") is None and "numberTitle" in values:
                values["uncountableTitle"] = values["numberTitle"]
        return values


class StepIngredientUnit(BaseModel):
    model_config = ConfigDict(alias_generator=to_camel)

    name: str | None = None
    localized_title: LocalizedStepIngredientUnit | None = None
    short_title: LocalizedStepIngredientUnit | None = None

    @model_validator(mode="before")
    def coerce_unit_string(cls, value: Any) -> dict[str, Any] | Any:
        if isinstance(value, str):
            return {"name": value}
        return value


class Ingredient(BaseModel):
    quantity: float | None = None
    measure: str | None = None
    ingredient: IngredientDetails


class StepIngredientDetails(BaseModel):
    model_config = ConfigDict(alias_generator=to_camel)

    id: RecipeId | None = Field(None, alias="_id")
    typ: str | None = None
    localized_title: LocalizedString | None = None
    number_title: LocalizedString | None = None
    uncountable_title: LocalizedString | None = None
    category: str | None = None

    @model_validator(mode="before")
    def coerce_titles(cls, values):
        if isinstance(values, str):
            return {"localizedTitle": values}
        if isinstance(values, dict):
            if values.get("localizedTitle") is None:
                for key in ("uncountableTitle", "title", "name", "numberTitle"):
                    if values.get(key) is not None:
                        values["localizedTitle"] = values[key]
                        break
            if (
                values.get("uncountableTitle") is None
                and values.get("numberTitle") is not None
            ):
                values["uncountableTitle"] = values["numberTitle"]
        return values


class StepIngredient(BaseModel):
    model_config = ConfigDict(alias_generator=to_camel)

    quantity: float | None = None
    unit: StepIngredientUnit | None = None
    ingredient: StepIngredientDetails | None = None
    ingredient_id: str | None = Field(None, alias="ingredientId")

    @model_validator(mode="before")
    def normalize_fields(cls, values):
        if isinstance(values, dict):
            if values.get("quantity") is None and values.get("amount") is not None:
                values["quantity"] = values["amount"]
            if values.get("unit") is None and values.get("measure") is not None:
                values["unit"] = values["measure"]
            if values.get("ingredientId") is None and isinstance(
                values.get("ingredient"), dict
            ):
                ingredient = values["ingredient"]
                ingredient_id = None
                if isinstance(ingredient.get("_id"), dict):
                    ingredient_id = ingredient["_id"].get("$oid")
                elif isinstance(ingredient.get("id"), dict):
                    ingredient_id = ingredient["id"].get("$oid")
                if ingredient_id is not None:
                    values["ingredientId"] = ingredient_id
        return values


class RecipeStep(BaseModel):
    title: LocalizedString
    image: Image
    ingredients: list[StepIngredient | None] | None = None


class Recipe(BaseModel):
    model_config = ConfigDict(alias_generator=to_camel)

    id: RecipeId = Field(..., alias="_id")
    uid: str | None = None
    rtype: str | None = None
    localized_title: LocalizedString
    author_comment: LocalizedString
    preparation_time: int
    cooking_time: int | None = None
    recipe_nutrition: Nutrition
    active_tags: list[str] | None = None
    steps: list[RecipeStep] = Field(..., alias="steps")
    image_list: list[Image]
    ingredients: list[Ingredient]

    @model_validator(mode="before")
    def normalize_titles(cls, values):
        if isinstance(values, dict):
            if values.get("localizedTitle") is None:
                title = values.get("title") or values.get("name")
                if title is not None:
                    values["localizedTitle"] = title
        return values

    def get_image_url(self, api_key: str) -> str | None:
        try:
            [image] = [i for i in self.image_list if i.type == "cover"]
        except ValueError:
            return None
        image_url = f"{image.url}?kptnkey={api_key}"
        return image_url



import datetime
import io
import json
import logging
import uuid
from getpass import getpass
from pathlib import Path
from typing import Any

import httpx
from pydantic import UUID4, BaseModel, ConfigDict, Field, ValidationError, parse_obj_as

from .config import settings
from .ingredient_groups import iter_ingredient_groups
from .models import (
    Image,
    Ingredient,
    RecipeStep as KptnCookRecipeStep,
    localized_fallback,
)
from .models import Recipe as KptnCookRecipe

logger = logging.getLogger(__name__)


class NameIsIdModel(BaseModel):
    model_config = ConfigDict(frozen=True)

    name: str

    def __hash__(self):
        """Hash on name to be able to have sets of models."""
        return hash(self.name)

    def __eq__(self, other):
        """Compare on name to be able to subtract sets of models."""
        return self.name == other.name


class RecipeTag(NameIsIdModel):
    slug: str | None = None
    name: str
    group_id: UUID4 | None = Field(None, alias="groupId")
    id: UUID4 | None = None


class RecipeCategory(RecipeTag):
    pass


class RecipeTool(RecipeTag):
    on_hand: bool = False


class UnitFoodBase(NameIsIdModel):
    id: UUID4 | None = None
    name: str
    description: str = ""


class RecipeFood(UnitFoodBase): ...  # noqa: E701


class RecipeUnit(UnitFoodBase):
    id: UUID4 | None = None
    fraction: bool = True
    abbreviation: str = ""


class RecipeIngredient(BaseModel):
    referenceId: UUID4 | None = None
    title: str | None = None
    note: str | None = None
    unit: RecipeUnit | None
    food: RecipeFood | None
    disable_amount: bool = True
    quantity: float | None = 1


class RecipeSummary(BaseModel):
    id: UUID4 | None = None

    user_id: UUID4 | None = Field(None, alias="userId")
    group_id: UUID4 | None = Field(None, alias="groupId")

    name: str | None = None
    slug: str = ""
    image: Any | None = None
    recipe_yield: str | None = None

    total_time: str | None = None
    prep_time: str | int | None = None
    cook_time: str | int | None = None
    perform_time: str | None = None

    description: str | None = ""
    recipe_category: list[str] | None = []
    tags: list[RecipeTag] | None = []
    # tags: list[RecipeTag | str] | None = []
    tools: list[RecipeTool] = []
    rating: int | None = None
    org_url: str | None = Field(None, alias="orgURL")

    recipe_ingredient: list[RecipeIngredient] | None = []

    date_added: datetime.date | None = None
    date_updated: datetime.datetime | None = None


class IngredientReference(BaseModel):
    referenceId: UUID4 | None = None


class RecipeStep(BaseModel):
    title: str | None = ""
    text: str
    ingredientReferences: list[IngredientReference] = []
    image: Image | None = None


class Nutrition(BaseModel):
    calories: str | None = None
    fatContent: str | None = None
    proteinContent: str | None = None
    carbohydrateContent: str | None = None
    fiberContent: str | None = None
    sodiumContent: str | None = None
    sugarContent: str | None = None


class RecipeSettings(BaseModel):
    public: bool = True
    show_nutrition: bool = True
    show_assets: bool = False
    landscape_view: bool = False
    disable_comments: bool = False
    disable_amount: bool = False
    locked: bool = False


class RecipeAsset(BaseModel):
    name: str
    icon: str
    file_name: str | None = None


class RecipeNote(BaseModel):
    title: str
    text: str | None = None


class Recipe(RecipeSummary):
    recipe_ingredient: list[RecipeIngredient] = []
    recipe_instructions: list[RecipeStep] | None = []
    nutrition: Nutrition | None = None

    # Mealie Specific
    settings: RecipeSettings | None = RecipeSettings()
    assets: list[RecipeAsset] | None = []
    notes: list[RecipeNote] | None = []
    extras: dict = {}


class RecipeWithImage(Recipe):
    image_url: str | None = None


class MealieApiClient:
    def __init__(self, base_url):
        self.base_url = base_url
        self.headers = {}

    @property
    def logged_in(self):
        return "access_token" in self.headers

    def to_url(self, path):
        return f"{self.base_url}{path}"

    def __getattr__(self, name):
        """
        Return proxy for httpx, joining base_url with path and
        providing authentication headers automatically.
        """

        def proxy(path, **kwargs):
            url = self.to_url(path)
            set_headers = kwargs.get("headers", {})
            kwargs["headers"] = set_headers | self.headers
            with httpx.Client() as client:
                response = getattr(client, name)(url, **kwargs)
            return response

        return proxy

    def fetch_api_token(self, username, password):
        login_data = {"username": username, "password": password}
        r = self.post("/auth/token", data=login_data, timeout=60)
        r.raise_for_status()
        return r.json()["access_token"]

    def login(self, username: str = "admin", password: str = ""):
        if password == "":
            password = getpass()
        access_token = self.fetch_api_token(username, password)
        self.headers = {"authorization": f"Bearer {access_token}"}

    def login_with_token(self, token: str):
        self.headers = {"authorization": f"Bearer {token}"}

    def upload_asset(self, recipe_slug, image: Image):
        # download image
        r = httpx.get(image.url, follow_redirects=True)
        r.raise_for_status()

        download = io.BytesIO(r.content)

        # upload to mealie
        name = Path(image.name)
        extension = name.suffix.lstrip(".")
        stem = name.stem
        data = {
            "name": stem,
            "icon": "mdi-file-image",
            "extension": extension,
        }
        r = self.post(
            f"/recipes/{recipe_slug}/assets", data=data, files={"file": download}
        )
        r.raise_for_status()

        return r.json()

    @staticmethod
    def _build_recipestep_text(recipe_uuid, text, image_name):
        return f'{text} <img src="/api/media/recipes/{recipe_uuid}/assets/{image_name}" height="100%" width="100%"/>'

    def enrich_recipe_with_step_images(self, recipe):
        assets = []
        for instruction in recipe.recipe_instructions:
            asset_properties = self.upload_asset(recipe.slug, instruction.image)
            uploaded_image_name = asset_properties["fileName"]
            instruction.text = self._build_recipestep_text(
                recipe.id, instruction.text, uploaded_image_name
            )
            assets.append(
                RecipeAsset(
                    name=asset_properties["name"],
                    icon=asset_properties["icon"],
                    file_name=asset_properties["fileName"],
                )
            )
        recipe.assets = assets
        return recipe

    def _post_recipe_trunk_and_get_slug(self, recipe_name):
        data = {"name": recipe_name}
        r = self.post("/recipes", data=json.dumps(data))
        r.raise_for_status()
        slug = r.json()
        return slug

    def _scrape_image_for_recipe(self, recipe, slug):
        if not recipe.image_url:
            return
        json_image_url = json.dumps({"url": recipe.image_url})
        scrape_image_path = f"/recipes/{slug}/image"
        r = self.post(scrape_image_path, data=json_image_url)
        r.raise_for_status()

    def _update_user_and_group_id(self, recipe, slug):
        recipe_detail_path = f"/recipes/{slug}"
        r = self.get(recipe_detail_path)
        r.raise_for_status()
        recipe_details = r.json()
        update_attributes = ["id", "userId", "groupId"]
        updated_details = {k: recipe_details[k] for k in update_attributes}
        recipe = RecipeWithImage(**(recipe.dict() | updated_details))
        return recipe

    def _get_page(self, endpoint_name, page_num, per_page=50):
        r = self.get(f"/{endpoint_name}?page={page_num}&perPage={per_page}")
        r.raise_for_status()
        return r.json()

    def _get_all_items(self, endpoint_name):
        all_items = []
        response_data = self._get_page(endpoint_name, 1)
        all_items.extend(response_data["items"])

        # 1 was already fetched, start page_num at 2 and add 1 to the
        # number of total pages, because we start counting at 1 instead of 0
        for page_num in range(2, response_data["total_pages"] + 1):
            response_data = self._get_page(endpoint_name, page_num)
            all_items.extend(response_data["items"])

        return all_items

    def _create_item(self, endpoint_name, item):
        r = self.post(f"/{endpoint_name}", data=item.json())
        r.raise_for_status()
        return r.json()

    def _create_item_name_to_item_lookup(self, endpoint_name, model_class, items):
        existing_items = parse_obj_as(
            set[model_class], self._get_all_items(endpoint_name)
        )
        items_to_create = items - existing_items
        for item in items_to_create:
            existing_items.add(model_class(**self._create_item(endpoint_name, item)))
        return {i.name: i for i in existing_items}

    def _update_item_ids(self, recipe, endpoint_name, model_class, attr_name):
        items = {
            getattr(ig, attr_name)
            for ig in recipe.recipe_ingredient
            if getattr(ig, attr_name) is not None
        }
        if len(items) == 0:
            # return early if there's nothing to do
            return recipe

        name_to_item_with_id = self._create_item_name_to_item_lookup(
            endpoint_name, model_class, items
        )
        for ingredient in recipe.recipe_ingredient:
            if getattr(ingredient, attr_name) is not None:
                setattr(
                    ingredient,
                    attr_name,
                    name_to_item_with_id[getattr(ingredient, attr_name).name],
                )
        return recipe

    def _update_tag_ids(self, recipe):
        recipe_tags = {tag for tag in recipe.tags}
        if len(recipe_tags) == 0:
            # return early if there's nothing to do
            return recipe

        name_to_tag_with_id = self._create_item_name_to_item_lookup(
            "organizers/tags", RecipeTag, recipe_tags
        )
        recipe.tags = [name_to_tag_with_id[tag.name] for tag in recipe_tags]
        return recipe

    def _update_recipe(self, recipe, slug):
        recipe_detail_path = f"/recipes/{slug}"
        r = self.put(recipe_detail_path, data=recipe.json())
        r.raise_for_status()
        return Recipe.model_validate(r.json())

    def create_recipe(self, recipe):
        slug = self._post_recipe_trunk_and_get_slug(recipe.name)
        logger.debug("Created Mealie recipe slug: %s", slug)
        recipe.slug = slug
        self._scrape_image_for_recipe(recipe, slug)
        recipe = self._update_user_and_group_id(recipe, slug)
        recipe = self._update_item_ids(recipe, "units", RecipeUnit, "unit")
        recipe = self._update_item_ids(recipe, "foods", RecipeFood, "food")
        recipe = self._update_tag_ids(recipe)
        recipe = self.enrich_recipe_with_step_images(recipe)
        return self._update_recipe(recipe, slug)

    @staticmethod
    def validate_recipes(recipes):
        validated_recipes = []
        for recipe in recipes:
            try:
                validated_recipe = Recipe.model_validate(recipe)
                validated_recipes.append(validated_recipe)
            except ValidationError as e:
                logger.error(
                    "Could not parse recipe {recipe_id}".format(recipe_id=recipe["id"])
                )
                logger.exception(e)
                continue
        return validated_recipes

    def get_all_recipes(self):
        all_recipes = []

        r = self.get("/recipes?page=1&perPage=50")
        r.raise_for_status()
        all_recipes.extend(self.validate_recipes(r.json()["items"]))

        page = 2
        while page <= r.json()["total_pages"]:
            r = self.get(f"/recipes?page={page}&perPage=50")
            r.raise_for_status()
            all_recipes.extend(self.validate_recipes(r.json()["items"]))
            page += 1

        return all_recipes

    def delete_via_slug(self, slug):
        r = self.delete(f"/recipes/{slug}")
        r.raise_for_status()
        return r.json()

    def get_via_slug(self, slug):
        r = self.get(f"/recipes/{slug}")
        r.raise_for_status()
        return Recipe.model_validate(r.json())


def kptncook_to_mealie_ingredients(
    kptncook_ingredients: list[Ingredient] | None,
) -> tuple[list[RecipeIngredient], dict[str, list[UUID4]]]:
    mealie_ingredients = []
    ingredient_id_to_reference_ids: dict[str, list[UUID4]] = {}
    groups = iter_ingredient_groups(kptncook_ingredients or [])
    for group_label, ingredients in groups:
        for ingredient in ingredients:
            ingredient_title = (
                localized_fallback(ingredient.ingredient.localized_title) or ""
            )
            note = None
            if "," in ingredient_title:
                ingredient_title, note, *_ = (
                    p.strip() for p in ingredient_title.split(",")
                )
            food = RecipeFood(name=ingredient_title)
            quantity = ingredient.quantity
            measure = None
            if hasattr(ingredient, "measure"):
                if ingredient.measure is not None:
                    measure = RecipeUnit(name=ingredient.measure)
            reference_id = uuid.uuid4()
            mealie_ingredient = RecipeIngredient(
                title=group_label or None,
                quantity=quantity,
                unit=measure,
                note=note,
                food=food,
                referenceId=reference_id,
            )
            mealie_ingredients.append(mealie_ingredient)
            ingredient_id = None
            if ingredient.ingredient.id is not None:
                ingredient_id = ingredient.ingredient.id.oid
            if ingredient_id:
                ingredient_id_to_reference_ids.setdefault(ingredient_id, []).append(
                    reference_id
                )
    return mealie_ingredients, ingredient_id_to_reference_ids


def kptncook_to_mealie_steps(
    steps: list[KptnCookRecipeStep],
    api_key: str,
    ingredient_id_to_reference_ids: dict[str, list[UUID4]],
):
    mealie_instructions = []
    for step in steps:
        image = step.image.get_image_with_api_key_url(api_key)
        reference_ids: list[UUID4] = []
        if step.ingredients:
            for ingredient in step.ingredients:
                if ingredient is None:
                    continue
                ingredient_id = ingredient.ingredient_id
                if (
                    ingredient_id is None
                    and ingredient.ingredient is not None
                    and ingredient.ingredient.id is not None
                ):
                    ingredient_id = ingredient.ingredient.id.oid
                if ingredient_id is None:
                    continue
                reference_ids.extend(
                    ingredient_id_to_reference_ids.get(ingredient_id, [])
                )
        seen = set()
        ingredient_references = []
        for reference_id in reference_ids:
            if reference_id in seen:
                continue
            seen.add(reference_id)
            ingredient_references.append(IngredientReference(referenceId=reference_id))
        mealie_instructions.append(
            RecipeStep(
                title=None,
                text=localized_fallback(step.title) or "",
                image=image,
                ingredientReferences=ingredient_references,
            )
        )
    return mealie_instructions


def kptncook_to_mealie_tags(active_tags: list[str] | None) -> list[RecipeTag]:
    tag_names = ["kptncook"]
    if active_tags:
        tag_names.extend(active_tags)

    seen = set()
    tags = []
    for name in tag_names:
        if not name or name in seen:
            continue
        seen.add(name)
        tags.append(RecipeTag.model_validate({"name": name, "group_id": None}))
    return tags


def kptncook_to_mealie(
    kcin: KptnCookRecipe, api_key: str = settings.kptncook_api_key
) -> RecipeWithImage:
    ingredients, ingredient_id_to_reference_ids = kptncook_to_mealie_ingredients(
        kcin.ingredients
    )
    kwargs = {
        "name": localized_fallback(kcin.localized_title),
        "notes": [
            RecipeNote(
                title="author comment",
                text=localized_fallback(kcin.author_comment),
            ),
        ],
        "nutrition": Nutrition(
            calories=str(kcin.recipe_nutrition.calories),
            proteinContent=str(kcin.recipe_nutrition.protein),
            fatContent=str(kcin.recipe_nutrition.fat),
            carbohydrateContent=str(kcin.recipe_nutrition.carbohydrate),
        ),
        "prep_time": kcin.preparation_time,
        "cook_time": kcin.cooking_time,
        "recipe_yield": "1 Portionen",
        "recipe_instructions": kptncook_to_mealie_steps(
            kcin.steps, api_key, ingredient_id_to_reference_ids
        ),
        "recipe_ingredient": ingredients,
        "image_url": kcin.get_image_url(api_key),
        "tags": kptncook_to_mealie_tags(kcin.active_tags),
        "extras": {"kptncook_id": kcin.id.oid, "source": "kptncook"},
    }
    logger.debug("Mealie recipe payload kwargs: %s", kwargs)
    return RecipeWithImage(**kwargs)
